/*************************************************************************\
 * Copyright (c) 2002 The University of Chicago, as Operator of Argonne
 * National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 * Operator of Los Alamos National Laboratory.
 * This file is distributed subject to a Software License Agreement found
 * in the file LICENSE that is included with this distribution. 
\*************************************************************************/

/* 
gcc -lgslcblas -lm  -lgsl sddskde.c
usage: sddskde input.sdds output.sdds -col=colData
author: Yipeng Sun, Hairong Shang
*/

#include <math.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_math.h>
#include "SDDS.h"
#include "mdb.h"
#include "scan.h"
#include "SDDSutils.h"

double bandwidth(double data[], int M);
double gaussiankernelfunction(double sample);
double kerneldensityestimate(double *trainingdata, double sample, int64_t n);
double *linearspace(double initial, double final, int N);

#define SET_COLUMN 0
#define SET_PIPE 1
#define SET_MARGIN 2
#define N_OPTIONS 3

char *option[N_OPTIONS] = {
  "column",
  "pipe",
  "margin",
};

static char *USAGE = "sddskde [<inputfile>] [<outputfile>] [-column=<list of columns>] [-pipe] [-margin=<value>]\n\
-column         provide column names separate by comma, wild card accepted.\n\
-margin         provide the ratio to extend the original data, default 0.3 \n\
-pipe           The standard SDDS Toolkit pipe option.\n\n\
sddskde does kernel density estimation for one dimensional data.\n\
Program by Yipeng Sun and Hairong Shang (" __DATE__ " " __TIME__ ", SVN revision: " SVN_VERSION ")\n";

int main(int argc, char **argv) {
  int n_test;
  double minTemp, maxTemp;
  double gap, lower, upper;
  double margin = 0.3;
  SDDS_DATASET SDDSin, SDDSout;
  char *inputFile, *outputFile, **column;
  long tmpfile_used = 0, columns, noWarnings = 1;
  unsigned long pipeFlags;
  long i, i_arg, j;
  SCANNED_ARG *s_arg;
  double *columnData, *pdf, *cdf;
  char buffer_pdf[1024], buffer_cdf[1024], buffer_pdfUnits[1024];
  int64_t rows;

  inputFile = outputFile = NULL;
  columnData = pdf = cdf = NULL;
  columns = 0;
  column = NULL;

  n_test = 100;
  margin = 0.3;

  SDDS_RegisterProgramName(argv[0]);
  argc = scanargs(&s_arg, argc, argv);
  pipeFlags = 0;
  tmpfile_used = 0;
  if (argc < 2) {
    fprintf(stderr, "%s", USAGE);
    exit(1);
  }
  for (i_arg = 1; i_arg < argc; i_arg++) {
    if (s_arg[i_arg].arg_type == OPTION) {
      delete_chars(s_arg[i_arg].list[0], "_");
      switch (match_string(s_arg[i_arg].list[0], option, N_OPTIONS, 0)) {
      case SET_COLUMN:
        columns = s_arg[i_arg].n_items - 1;
        column = trealloc(column, sizeof(*column) * columns);
        for (i = 1; i < s_arg[i_arg].n_items; i++)
          column[i - 1] = s_arg[i_arg].list[i];
        break;
      case SET_MARGIN:
        if (s_arg[i_arg].n_items != 2)
          SDDS_Bomb("Invalid -margin option!");
        if (!get_double(&margin, s_arg[i_arg].list[1]))
          SDDS_Bomb("Invalid -margin value provided.!");
        break;
      case SET_PIPE:
        if (!processPipeOption(s_arg[i_arg].list + 1, s_arg[i_arg].n_items - 1, &pipeFlags)) {
          fprintf(stderr, "Error (%s): invalid -pipe syntax\n", argv[0]);
          return (1);
        }
        break;
      }
    } else {
      if (!inputFile)
        inputFile = s_arg[i_arg].list[0];
      else if (outputFile == NULL)
        outputFile = s_arg[i_arg].list[0];
      else {
        fprintf(stderr, "Error (%s): too many filenames\n", argv[0]);
        return (1);
      }
    }
  }
  processFilenames("sddskde", &inputFile, &outputFile, pipeFlags, noWarnings, &tmpfile_used);
  if (!columns) {
    fprintf(stderr, "%s", USAGE);
    SDDS_Bomb("No column provided!");
  }
  if (!SDDS_InitializeInput(&SDDSin, inputFile)) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    return (1);
  }
  if ((columns = expandColumnPairNames(&SDDSin, &column, NULL, columns, NULL, 0, FIND_NUMERIC_TYPE, 0)) <= 0) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    SDDS_Bomb("no columns selected.");
  }
  if (!SDDS_InitializeOutput(&SDDSout, SDDS_BINARY, 1, NULL, NULL, outputFile))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  for (i = 0; i < columns; i++) {
    char *units;
    if (SDDS_GetColumnInformation(&SDDSin, "units", &units, SDDS_GET_BY_NAME, column[i]) != SDDS_STRING)
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    if (units && strlen(units))
      snprintf(buffer_pdfUnits, 1024, "1/(%s)", units);
    else
      buffer_pdfUnits[0] = 0;
    free(units);
    units = NULL;
    snprintf(buffer_pdf, 1024, "%sPDF", column[i]);
    snprintf(buffer_cdf, 1024, "%sCDF", column[i]);
    if (!SDDS_TransferColumnDefinition(&SDDSout, &SDDSin, column[i], NULL) ||
        SDDS_DefineColumn(&SDDSout, buffer_pdf, NULL, buffer_pdfUnits, NULL, NULL, SDDS_DOUBLE, 0) < 0 ||
        SDDS_DefineColumn(&SDDSout, buffer_cdf, NULL, NULL, NULL, NULL, SDDS_DOUBLE, 0) < 0)
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  }
  if (!SDDS_WriteLayout(&SDDSout))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  pdf = malloc(sizeof(*pdf) * n_test);
  cdf = malloc(sizeof(*cdf) * n_test);
  while (SDDS_ReadPage(&SDDSin) >= 0) {
    if ((rows = SDDS_CountRowsOfInterest(&SDDSin))) {
      if (!SDDS_StartPage(&SDDSout, n_test))
        SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
      for (i = 0; i < columns; i++) {
        sprintf(buffer_pdf, "%sPDF", column[i]);
        sprintf(buffer_cdf, "%sCDF", column[i]);
        if (!(columnData = SDDS_GetColumnInDoubles(&SDDSin, column[i])))
          SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
        find_min_max(&minTemp, &maxTemp, columnData, rows);
        gap = maxTemp - minTemp;
        lower = minTemp - gap * margin;
        upper = maxTemp + gap * margin;
        double *x_array = linearspace(lower, upper, n_test);
        for (j = 0; j < n_test; j++) {
          pdf[j] = kerneldensityestimate(columnData, x_array[j], rows);
          cdf[j] = pdf[j];
        }
        for (j = 1; j < n_test; j++)
          cdf[j] += cdf[j - 1];
        /* cumsumrec(cdf n_test-1); */
        for (j = 0; j < n_test; j++) {
          cdf[j] = cdf[j] / cdf[n_test - 1];
        }
        if (!SDDS_SetColumnFromDoubles(&SDDSout, SDDS_SET_BY_NAME, pdf, n_test, buffer_pdf) ||
            !SDDS_SetColumnFromDoubles(&SDDSout, SDDS_SET_BY_NAME, cdf, n_test, buffer_cdf) ||
            !SDDS_SetColumnFromDoubles(&SDDSout, SDDS_SET_BY_NAME, x_array, n_test, column[i]))
          SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
        free(columnData);
        columnData = NULL;
        free(x_array);
        x_array = NULL;
      }
    }
    if (!SDDS_WritePage(&SDDSout))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  }
  free(pdf);
  free(cdf);
  if (!SDDS_Terminate(&SDDSin) || !SDDS_Terminate(&SDDSout))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  if (tmpfile_used && !replaceFileAndBackUp(inputFile, outputFile)) {
    return (1);
  }
  free(column);
  /*  free_scanargs(&s_arg,argc);  results in memory error from valgrind --tool=memcheck */
  return 0;
}

double bandwidth(double data[], int M) {
  double sigma, min, bwidth, interquartile_range;
  double hspread_three, hspread_one;
  gsl_sort(data, 1, M);
  sigma = gsl_stats_sd(data, 1, M);
  hspread_three = gsl_stats_quantile_from_sorted_data(data, 1, M, 0.750);
  hspread_one = gsl_stats_quantile_from_sorted_data(data, 1, M, 0.250);
  interquartile_range = hspread_three - hspread_one;
  min = GSL_MIN(interquartile_range / 1.339999, sigma);
  bwidth = 0.90 * min * pow(M, -0.20);
  return bwidth;
}

double gaussiankernelfunction(double sample) {
  double k;
  k = exp(-(gsl_pow_2(sample) / 2.0));
  k = k / (M_SQRT2 * sqrt(M_PI));
  return k;
}

double kerneldensityestimate(double *trainingdata, double sample, int64_t n) {
  int64_t i;
  double pdf, h, bwidth;

  bwidth = bandwidth(trainingdata, n);
  h = GSL_MAX(bwidth, 2e-6);
  pdf = 0.0;
  for (i = 0; i < n; i++) {
    pdf += gaussiankernelfunction((trainingdata[i] - sample) / h);
  }
  pdf = pdf / (n * h);
  return pdf;
}

double *linearspace(double start, double end, int N) {
  double *x;
  int i;
  double step;

  x = (double *)calloc(N, sizeof(double));
  step = (end - start) / (double)(N - 1);

  i = 0;
  x[i] = start;
  for (i = 1; i < N; i++) {
    x[i] = x[i - 1] + step;
  }
  x[N - 1] = end;
  return x;
}
