
#include "gnuplot_plugin.h"
#include "SDDS.h"

SDDS_DATASET SDDS_dataset;

DLLEXPORT struct value initInput(int nargs, struct value *arg, void *p)
{
  char *filename = arg[0].v.string_val;
  struct value r;

  /* Enforce a match between the number of parameters declared
   * by the gnuplot import command and the number implemented here.
   */
  RETURN_ERROR_IF_WRONG_NARGS(r, nargs, 1);

  if (SDDS_InitializeInput(&SDDS_dataset, filename) == 0) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    exit(1); //Locks up if I don't force an exit here
    //r.type = INVALID_VALUE;
    //return r;
  }
  r.type = INTGR;
  r.v.int_val = 1;
  return r;
}

DLLEXPORT struct value closeInput(int nargs, struct value *arg, void *p)
{
  struct value r;

  if (SDDS_Terminate(&SDDS_dataset) == 0) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    exit(1);
  }
  r.type = INTGR;
  r.v.int_val = 1;
  return r;
}

DLLEXPORT struct value columnCount(int nargs, struct value *arg, void *p)
{
  struct value r;
  int count;

  count = SDDS_ColumnCount(&SDDS_dataset);
  if (count < 0) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    exit(1);
  }
  r.type = INTGR;
  r.v.int_val = count;
  return r;
}

DLLEXPORT struct value columnName(int nargs, struct value *arg, void *p)
{
  int index = IVAL(arg[0]);
  struct value r;
  int32_t number;
  char **names;

  RETURN_ERROR_IF_WRONG_NARGS(r, nargs, 1);
  RETURN_ERROR_IF_NONNUMERIC(r, arg[0]);
  names = SDDS_GetColumnNames(&SDDS_dataset, &number);

  r.type = STRING;
  r.v.string_val = names[index];
  return r;
}

DLLEXPORT struct value readPage(int nargs, struct value *arg, void *p)
{
  int page = IVAL(arg[0]);
  struct value r;
  int pageRead;

  /* Enforce a match between the number of parameters declared
   * by the gnuplot import command and the number implemented here.
   */
  RETURN_ERROR_IF_WRONG_NARGS(r, nargs, 1);

  /* Sanity check on argument type */
  RETURN_ERROR_IF_NONNUMERIC(r, arg[0]);

  pageRead = SDDS_ReadPage(&SDDS_dataset);
  while (pageRead < page) {
    if (pageRead==0) {
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
      r.type = INVALID_VALUE;
      return r;
    }
    if (pageRead==-1) {/*EOF*/
      fprintf(stderr, "EOF reached in SDDS file\n");
      r.type = INVALID_VALUE;
      return r;
    }
    pageRead = SDDS_ReadPage(&SDDS_dataset);
  }
  r.type = INTGR;
  r.v.int_val = SDDS_RowCount(&SDDS_dataset);
  return r;
}

DLLEXPORT struct value columnIndex(int nargs, struct value *arg, void *p)
{
  char *name = arg[0].v.string_val;
  struct value r;
  int index;

  /* Enforce a match between the number of parameters declared
   * by the gnuplot import command and the number implemented here.
   */
  RETURN_ERROR_IF_WRONG_NARGS(r, nargs, 1);

  index = SDDS_GetColumnIndex(&SDDS_dataset, name);
  if (index < 0) {
    fprintf(stderr, "Column %s not found in SDDS file\n", name);
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    r.type = INVALID_VALUE;                     
    return r;
  }
  r.type = INTGR;
  r.v.int_val = index;

  return r;
}

DLLEXPORT struct value rowValue(int nargs, struct value *arg, void *p)
{
  char *name = arg[0].v.string_val;
  int row = IVAL(arg[1]);
  double value;
  struct value r;
  r.type = CMPLX;

  /* Enforce a match between the number of parameters declared
   * by the gnuplot import command and the number implemented here.
   */
  RETURN_ERROR_IF_WRONG_NARGS(r, nargs, 2);

  /* Sanity check on argument type */
  RETURN_ERROR_IF_NONNUMERIC(r, arg[1]);

  value = SDDS_GetValueAsDouble(&SDDS_dataset, name, row-1);
  
  r.v.cmplx_val.real = value;
  r.v.cmplx_val.imag = 0.0;

  return r;
}

DLLEXPORT struct value rowValueByIndex(int nargs, struct value *arg, void *p)
{
  int index = IVAL(arg[0]);
  int row = IVAL(arg[1]);
  double value;
  struct value r;
  r.type = CMPLX;

  /* Enforce a match between the number of parameters declared
   * by the gnuplot import command and the number implemented here.
   */
  RETURN_ERROR_IF_WRONG_NARGS(r, nargs, 2);

  /* Sanity check on argument type */
  RETURN_ERROR_IF_NONNUMERIC(r, arg[0]);
  RETURN_ERROR_IF_NONNUMERIC(r, arg[1]);

  value = SDDS_GetValueByIndexAsDouble(&SDDS_dataset, index, row-1);
  
  r.v.cmplx_val.real = value;
  r.v.cmplx_val.imag = 0.0;

  return r;
}

DLLEXPORT struct value parameterValue(int nargs, struct value *arg, void *p)
{
  char *name = arg[0].v.string_val;
  double *value=NULL;
  struct value r;
  r.type = CMPLX;

  /* Enforce a match between the number of parameters declared
   * by the gnuplot import command and the number implemented here.
   */
  RETURN_ERROR_IF_WRONG_NARGS(r, nargs, 1);

  /* Sanity check on argument type */
  RETURN_ERROR_IF_NONNUMERIC(r, arg[1]);

  value = SDDS_GetParameterAsDouble(&SDDS_dataset, name, value);
  if (value == NULL) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    r.type = INVALID_VALUE;
    return r;
  }
  r.v.cmplx_val.real = *value;
  r.v.cmplx_val.imag = 0.0;

  free(value);

  return r;
}

DLLEXPORT struct value parameterValueString(int nargs, struct value *arg, void *p)
{
  char *name = arg[0].v.string_val;
  char *value=NULL;
  struct value r;
  r.type = STRING;

  /* Enforce a match between the number of parameters declared
   * by the gnuplot import command and the number implemented here.
   */
  RETURN_ERROR_IF_WRONG_NARGS(r, nargs, 1);

  /* Sanity check on argument type */
  RETURN_ERROR_IF_NONNUMERIC(r, arg[1]);

  r.v.string_val = NULL;
  r.v.string_val = SDDS_GetParameterAsString(&SDDS_dataset, name, &r.v.string_val);
  if (r.v.string_val == NULL) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    r.type = INVALID_VALUE;                     \
    return r;                                   \
  }

  return r;
}

